#include "formule.hpp"

// BEGIN Création de formules simples //////////////////////////////////////////
formule variable(string nom) {
  formule res = new formule_s;
  res->op = o_variable;
  res->nom = new string(nom);
  return res;
}

formule variable(char* nom) {
  formule res = new formule_s;
  res->op = o_variable;
  res->nom = new string(nom);
  return res;
}

formule non(formule arg) {
  formule res = new formule_s;
  res->op = o_non;
  res->arg = arg;
  return res;
}

formule et(formule arg1, formule arg2) {
  formule res = new formule_s;
  res->op = o_et;
  res->arg1 = arg1;
  res->arg2 = arg2;
  return res;
}

formule ou(formule arg1, formule arg2) {
  formule res = new formule_s;
  res->op = o_ou;
  res->arg1 = arg1;
  res->arg2 = arg2;
  return res;
}

formule impl(formule arg1, formule arg2) {
  formule res = new formule_s;
  res->op = o_implique;
  res->arg1 = arg1;
  res->arg2 = arg2;
  return res;
}

formule equiv(formule arg1, formule arg2) {
  formule res = new formule_s;
  res->op = o_equivaut;
  res->arg1 = arg1;
  res->arg2 = arg2;
  return res;
}
// END Création de formules simples ////////////////////////////////////////////


// BEGIN Affichage d'une formule ///////////////////////////////////////////////
// Affichage des opérateurs
string operateur2string(operateur op) {
  switch (op) {
  case o_variable:
    return "var";
  case o_non:
    return "~";
  case o_et:
    return "/\\";
  case o_ou:
    return "\\/";
  case o_implique:
    return "=>";
  case o_equivaut:
    return "<=>";
  }
  return string();
}

// Affichage d'une formule par induction sur son arbre de syntaxe
string formule2string(formule f) {
  switch (f->op) {
  case o_variable: 
    return *(f->nom);
  case o_non:
    return "~" + formule2string(f->arg);
  case o_et:
  case o_ou:
  case o_equivaut:
  case o_implique:
    return "("+formule2string(f->arg1)
      +" "+operateur2string(f->op)+" "
      +formule2string(f->arg2)+")";
  }
  return string();
}

// Définition de l'opérateur << pour le type formule
ostream& operator<<(ostream& out, formule f){
  out << formule2string(f);
  return out;
}
// END Affichage d'une formule /////////////////////////////////////////////////


// BEGIN Gestion de la mémoire dynamique des formules //////////////////////////
formule copier(formule f) {
  switch (f->op) {
  case o_variable:
    return variable(* f->nom);
  case o_non:
    return non(copier(f->arg));
  case o_et:
    return et(copier(f->arg1),copier(f->arg2));
  case o_ou:
    return ou(copier(f->arg1),copier(f->arg2));
  case o_implique:
    return impl(copier(f->arg1),copier(f->arg2));
  case o_equivaut:
    return equiv(copier(f->arg1),copier(f->arg2));
  }
  return nullptr;
}

void free_all(formule f) {
  switch (f->op) {
  case o_implique:
  case o_equivaut:
  case o_ou:
  case o_et:
    free_all(f->arg1);
    free_all(f->arg2);
    break;
  case o_non:
    free_all(f->arg);
    break;
  case o_variable:
    delete f->nom;
    break;
  }
  delete f;
}

// END Gestion de la mémoire dynamique des formules ////////////////////////////


////////////////////////////////////////////////////////////////////////////////
//            A PARTIR D'ICI LES FONCTIONS SONT A IMPLEMENTER                 //
////////////////////////////////////////////////////////////////////////////////
ostream& operator<<(ostream& out, const cls_t& cl){
  // A FAIRE
  // exemple 
  out << "{" << 42 << '}' << endl;
  return out;
}

ostream& operator<<(ostream& out, const cnf_t& c){
  // A FAIRE
  return out;
}


bool est_dans(const cls_t& c, const cnf_t& f){
  // A FAIRE
  return false;
}

formule applique(formule f, const subs_t & sub) {
  // A FAIRE
  return NULL;
}

void numerote(formule f, map<string,var_t> & corresp){
  // A FAIRE
}

lit_t var2lit(var_t v, bool p){
  // A FAIRE
  return 0;
}

lit_t neg(lit_t l){
  // A FAIRE
  return 0;
}

var_t lit2var(lit_t l) {
  // A FAIRE
  return 0;
}

bool positif(lit_t l) {
  // A FAIRE
  return true;
}

lit_t tseitin(formule f, map<string,var_t> & corresp, cnf_t& cnf, var_t & next) {
  // A FAIRE
  return 0;
}

void tseitin(formule f, map<string,var_t> & corresp, cnf_t& cnf) {
  // A FAIRE
}

////////////////////////////////////////////////////////////////////////////////
// Fonctions annexes fournies
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Ecriture au format dimacs attendu par minisat
void writeDimacs(ostream&out, const map<string,int> & corresp, const cnf_t& cnf) {
  out << "c Fichier genere" << endl;
  out << "p cnf " << corresp.size() << " " << cnf.size() << endl;
  for(auto cnf_it = cnf.cbegin(); cnf_it != cnf.cend(); ++cnf_it) {
    for(auto cl_it = (*cnf_it).cbegin(); cl_it != (*cnf_it).cend(); ++cl_it) {
      lit_t l = *cl_it;
      out << ( (positif(l)?1:-1) * (1+lit2var(l)) ) << " ";
    }
    out << 0 << endl;
  }
}

////////////////////////////////////////////////////////////////////////////////
// Code pour la réécriture en CNF équivalente d'une formule
cnf_t cnf_union(const cnf_t& lhs, const cnf_t& rhs){
  cnf_t res = lhs;
  res.reserve(lhs.size() + rhs.size());
  res.insert(res.end(), rhs.cbegin(), rhs.cend());  
  return res;
}

cnf_t cnf_product(const cnf_t& lhs, const cnf_t& rhs){
//product E x F = { (x union y) | x in E /\ y in F}
  cnf_t res;
  cls_t xy;
  for (auto& x : lhs){
      for (auto& y : rhs){
        xy.clear();
        set_union(x.cbegin(), x.cend(), y.cbegin(), y.cend(), std::inserter(xy, xy.end()));//
        res.push_back(xy);
      }
  }
  return res;
}

// Déclaration pour les appels mutuellement recursifs
cnf_t rewriteDNF(formule f, const map<string,var_t> & corresp, bool pos=true);

cnf_t rewriteCNF(formule f, const map<string,var_t> & c, bool pos) {
  cnf_t res;
  switch (f->op) {
    case o_variable:{
      res = { {var2lit(c.find(* (f->nom))->second, pos)} };
      return (res);
      }
    case o_non:
      return rewriteDNF(f->arg, c, !pos);
      break;
    case o_et: {
      cnf_t c1 = rewriteCNF(f->arg1, c, pos);
      cnf_t c2 = rewriteCNF(f->arg2, c, pos);
      return cnf_union(c1, c2);
      }
    case o_ou: {
      cnf_t c1 = rewriteCNF(f->arg1, c, pos);
      cnf_t c2 = rewriteCNF(f->arg2, c, pos);
      return cnf_product(c1, c2);
      } 
    case o_implique: {
      cnf_t c1 = rewriteDNF(f->arg1, c, !pos);
      cnf_t c2 = rewriteCNF(f->arg2, c, pos);
      return cnf_product(c1, c2);
    }
    case o_equivaut: {
      cnf_t c1a = rewriteDNF(f->arg1, c, !pos);
      cnf_t c2a = rewriteCNF(f->arg2, c, pos);
      cnf_t ca = cnf_product(c1a, c2a);
      cnf_t c1b = rewriteCNF(f->arg1, c, !pos);
      cnf_t c2b = rewriteDNF(f->arg2, c, pos);
      cnf_t cb = cnf_product(c1b, c2b);
      return cnf_union(ca, cb);
      }
    }
}


cnf_t rewriteDNF(formule f, const map<string,var_t> & c, bool pos) {
  cnf_t res;
  switch (f->op) {
    case o_variable:{
      res = { {var2lit(c.find(* (f->nom))->second, pos)} };
      return (res);
      }
    case o_non:
      return rewriteCNF(f->arg, c, !pos);
      break;
    case o_et: {
      cnf_t c1 = rewriteDNF(f->arg1, c, pos);
      cnf_t c2 = rewriteDNF(f->arg2, c, pos);
      return cnf_product(c1, c2);
      }
    case o_ou: {
      cnf_t c1 = rewriteDNF(f->arg1, c, pos);
      cnf_t c2 = rewriteDNF(f->arg2, c, pos);
      return cnf_union(c1, c2);
      } 
    case o_implique: {
      cnf_t c1 = rewriteCNF(f->arg1, c, !pos);
      cnf_t c2 = rewriteDNF(f->arg2, c, pos);
      return cnf_union(c1, c2);
    }
    case o_equivaut: {
      cnf_t c1a = rewriteCNF(f->arg1, c, !pos);
      cnf_t c2a = rewriteDNF(f->arg2, c, pos);
      cnf_t ca = cnf_union(c1a, c2a);
      cnf_t c1b = rewriteDNF(f->arg1, c, !pos);
      cnf_t c2b = rewriteCNF(f->arg2, c, pos);
      cnf_t cb = cnf_union(c1b, c2b);
      return cnf_product(ca, cb);
      }
    }
}

